package loopbounds;

import java.util.List;
import java.util.ArrayList;

import org.jmlspecs.openjml.JmlTree;

import com.sun.tools.javac.tree.JCTree.*;
import com.sun.tools.javac.util.Context;

/**
 * Class to store a piecewise polynomial loop-bound function.
 * 
 * @author Rody Kersten
 * 
 */
public class PiecewisePolynomial {
	private List<JCExpression> conditions;
	private List<Polynomial> polynomials;
	
	/** Create a piecewise polynomial from a list of conditions and an equally long
	 * list of polynomials.
	 * 
	 * @param conditions List of conditions of length n.
	 * @param polynomials List of polynomials of length n.
	 */
	public PiecewisePolynomial(List<JCExpression> conditions, List<Polynomial> polynomials) {
		assert(conditions.size() == polynomials.size());
		this.conditions = conditions;
		this.polynomials = polynomials;
	}
	
	/** Create a piecewise polynomial. */
	public PiecewisePolynomial() {
		this(new ArrayList<JCExpression>(), new ArrayList<Polynomial>());
	}
	
	/** Create a piecewise polynomial containing one initial piece.
	 * 
	 * @param condition Condition
	 * @param polynomial Polynomial
	 */
	public PiecewisePolynomial(JCExpression condition, Polynomial polynomial) {
		this();
		add(condition,polynomial);
	}
	
	/** Add a piece.
	 * 
	 * @param condition Condition of the piece to add.
	 * @param polynomial Polynomial of the piece to add.
	 * @return True if both additions were successful.
	 */
	public boolean add(JCExpression condition, Polynomial polynomial) {
		boolean bc = conditions.add(condition);
		boolean bp = polynomials.add(polynomial);
		return bc&&bp;
	}
	
	/** Check if this Piecewise Polynomial is undefined. Used if a LBF cannot be found.
	 * 
	 * @return True if this piecewise polynomial is undefined
	 */
	public boolean isUndefined() {
		return conditions.isEmpty() || polynomials.isEmpty();
	}

	/** Export this piecewise polynomial as a JCExpression.
	 * 
	 * @param context The OpenJML context
	 * @return The piecewise polynomial LBF as a JCExpression.
	 */
	public JCExpression asDecreasesClause(Context context) {
		assert(conditions.size() == polynomials.size());
		JmlTree.Maker M = JmlTree.Maker.instance(context);
		
		JCExpression decr = M.Literal(0);
		
		for (int i = conditions.size()-1; i >= 0; i--) {
			decr = M.Conditional(conditions.get(i), polynomials.get(i).toJCExpression(context), decr);
		}
		
		return decr;
	}
	
	/** Get the number of pieces of this piecewise polynomial.
	 * 
	 * @return The number of pieces in this piecewise polynomial
	 */
	public int numPieces() {
		assert(conditions.size() == polynomials.size());
		return conditions.size();
	}
	
	/** Get all the variables used in this piecewise polynomial.
	 * Look in all the pieces which variables are used and return a list
	 * of the unique variables.
	 * 
	 * @return A list of all the variables used in the piecewise polynomial
	 */
	public List<String> getUniqueVariables() {
		List<String> vars = new ArrayList<String>();
		for(Polynomial p : polynomials) {
			for (String var : p.getVariables().getFullNames()) {
				if (!vars.contains(var))
					vars.add(var);
			}
		}
		return vars;
	}
}
