package loopbounds.ng;

import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import loopbounds.LoopBoundsAnalyzer;

import org.codehaus.janino.ExpressionEvaluator;

import com.sun.tools.javac.tree.JCTree.JCExpression;


/**
 * Class for generating the test nodes for polynomial interpolation. The constructor takes
 * as arguments the degree, the number of variables, and a ConditionEvaluator. Only nodes
 * that satisfy the given condition will be produced.
 * 
 * Starts with the 'startPoint', then iteratively adds 'step' in both directions on each
 * axis and checks whether these nodes are OK. Also transposes the startPoint on each
 * iteration.
 * 
 * @author Rody Kersten
 *
 */
public class SimpleNodeGenerator implements NodeGenerator {
	private int d;	//degree
	private int step; //step
	private String argNames[]; //variables
	private Class<?>[] argTypes; //variable types
	
	private JCExpression condition;
	private ExpressionEvaluator ee;
	private List<int[]> nodes;
	
	//startpoint and bounds for the search
	private int startPoint;
	private int minval;
	private int maxval;
	
	/** Constructor.
	 * 
	 * @param d The expected degree
	 * @param step The step
	 * @param argNames The names of the variables
	 * @param argTypes The types of the variables
	 * @param condition The loop condition (over the aforementioned variables)
	 * @param startPoint Startpoint for the nodesearch
	 * @param minval Minimal bound above which nodes are searched
	 * @param maxval Maximal bound below which nodes are searched
	 * @throws Exception
	 */
	public SimpleNodeGenerator (int d, int step, String[] argNames, Class<?>[] argTypes,
			JCExpression condition, int startPoint, int minval, int maxval)
		throws org.codehaus.janino.Parser.ParseException, org.codehaus.janino.CompileException, org.codehaus.janino.Scanner.ScanException {

		this.d = d;
		this.step = step;
		this.argNames = argNames;
		this.argTypes = argTypes;
		this.nodes = new ArrayList<int[]>();
		this.condition = condition;
		this.ee = new ExpressionEvaluator(condition.toString(), boolean.class, argNames, argTypes);
		this.startPoint = startPoint;
		this.minval = minval;
		this.maxval = maxval;
	}
	
	/**
	 * Implements a simple algorithm to find test nodes.
	 * Find sufficient nodes within a window that increases its size.
	 * Finds nodes that are 'step' apart. Transposes the start position
	 * on each round, so that nodes are also found if the don't lie n*step
	 * from the root.
	 * 
	 * The found test nodes satisfy both the NCA configuration and
	 * the loop condition. Throws an exception when these nodes can
	 * not be found.
	 */
	public void generate() throws NotEnoughTestNodesException, InvocationTargetException {
		
		//Special case: if there are no variables, we need no test-nodes
		if (argNames == null || argNames.length == 0) {
			nodes.add(new int[0]);
			return;
		}
			
		//size of the considered cube (in steps)
		int length = d;
		
		//transpose the cube on each recursive round
		int transpose = 0;
		
		while(!done()) {

			if (startPoint+length*step > this.maxval
					|| startPoint-length*step < this.minval) {
				//If all possible nodes have been tried,
				//i.e. the length is larger than the largest window dimension
				//throw an exception, because it is then impossible to interpolate
				throw new NotEnoughTestNodesException(condition.toString(), minval, maxval);
			}
			
			//generate nodes for current length
			nodes.clear();
			generateAux(length, transpose, null);
			
			if (LoopBoundsAnalyzer.DEBUG)
				System.out.println("Length = " + length);
			
			//increase length
			length *= 2;
			
			//transpose
			transpose = transpose + 1 % step;
		}
		
		//Reduce to minimum set
		reduceToSquare();
	}
	
	/** Recursive on the considered dimension. Is called initially from generate()
	 *  on the first dimension. Iteratively fixes the value for each dimension.
	 *  
	 * @param length The size of the considered cube.
	 * @param transpose The transposition of the cube.
	 * @param fixed The fixed values for preceding dimensions
	 * 				(i.e. the part of the node that is fixed thus far).
	 * 				Can be null if no dimensions have been considered yet.
	 * @throws InvocationTargetException
	 */
	private void generateAux(int length, int transpose, int[] fixed) throws InvocationTargetException {
		
		//The dimension we look at is the first one that hasn't been done yet
		int dim = (fixed==null) ? 0 : fixed.length;
		if (LoopBoundsAnalyzer.DEBUG)
			System.out.println("generateAux called with dim="+dim);
		
		//Check if this is the last dimension
		boolean lastDim = (dim == argNames.length-1);
		
		//new version that looks in both directions:
		for (int dist = 0; dist <= length*step; dist += step) {

			//for direction = -1 and direction = 1
			for (int direction = -1; direction <= 1; direction+=2) {
				int val = startPoint + transpose + direction*dist; // look in 2 directions

		//old version that did not look in the negative direction:
		//		for (int val = transpose; val <= transpose+length*step; val += step) {

				if (val < this.minval || val > this.maxval)
					continue;

				//fix the value for this dimension
				int newFixed[] = new int[dim+1];
				if (dim > 0) {
					for (int i = 0; i < dim; i++)
						newFixed[i] = fixed[i];
				}
				newFixed[dim] = val;

				if (lastDim) {
					if (LoopBoundsAnalyzer.DEBUG)
						System.out.println("Trying node: " + Arrays.toString(newFixed));

					//Transform int[] into Object[] for evaluation
					Object[] values = new Object[argNames.length];
					for (int i = 0; i < argNames.length; i++) {
                        // Generate the correct class as argument instead of casting it to Integer
                        // The loop condition evaluator will fail if the type is not given correctly
					    Object o;
                        if (argTypes[i].equals(byte.class) || argTypes[i].equals(Byte.class)) {
                            o = Byte.valueOf((byte) newFixed[i]);
                        } else if (argTypes[i].equals(short.class) || argTypes[i].equals(Short.class)) {
                            o = Short.valueOf((short) newFixed[i]);
                        }else if (argTypes[i].equals(long.class) || argTypes[i].equals(Long.class)) {
                            o = Long.valueOf( newFixed[i]);
                        } else {
                            o = newFixed[i];
                        }
						if (argTypes[i].isArray()) {
							values[i] = Array.newInstance(argTypes[i].getComponentType(), newFixed[i]+1);
							for (int j = 0; j < newFixed[i]; j++) { // Try to fill it with values 
								Array.set(values[i], j, o);
							}
						} else {
						    values[i] = o;
						}
					}


					//This is the last dimension, i.e. the entire node is fixed.
					//Evaluate the node, if it's not already part of 'nodes'.
					if (!alreadyTried(newFixed) && (Boolean) ee.evaluate(values))
						nodes.add(newFixed);
				} else {
					//There are more dimensions that aren't fixed yet, recurse...
					generateAux(length, transpose, newFixed);
				}
			}
		}
	}
	
	/** Checks if we have already considered this node **/
	private boolean alreadyTried(int node[]) {
		for (int goodnode[] : nodes) {
			if (Arrays.equals(node, goodnode))
				return true;
		}
		return false;
	}
	
	/** Get the number of test nodes found.
	 * 
	 * @return The number of nodes found.
	 */
	public int getNumberOfNodes () {
		return this.nodes.size();
	}
	
	/** Get the nodes.
	 * 
	 * @return The found test nodes (after 'generate' has been called), otherwise an empty array.
	 */
	public List<int[]> getNodes () {
		return this.nodes;
	}
	
	/** Check whether we have found enough nodes on enough separate
	 * hyperplanes to interpolate a polynomial.
	 * 
	 * When d is our degree, there should be at least d+1 hyperplanes,
	 * where one hyperplane contains at least d+1 nodes, another contains
	 * at least d nodes, ..., and the last contains at least 1 node.
	 * 
	 * @return True if we have found enough nodes
	 */
	private boolean done() {
		return inNCA(d+1, nodes, argNames);
	}
	
	private boolean inNCA(int triangleSize, List<int[]> nodes, String[] vars) {
		
		if (LoopBoundsAnalyzer.DEBUG)
			System.out.println("inNCA() called with arguments: size=" + triangleSize + ", nodes(length)=" + nodes.size() + ", vars(length)=" + vars.length);
			
		//if there is only one variable/dimension left, then there are enough
		//nodes if there are as least as many nodes as the triangleSize
		if (vars.length == 1) {
			boolean ret = (nodes.size() >= triangleSize);
			if (LoopBoundsAnalyzer.DEBUG)
				System.out.println("Returning " + ret + " for vars.size()==1");
			return ret;
		}
		
		//There is a minimum number of nodes that is needed to interpolate.
		//For efficiency, quit here if there are too few.
		if (nodes.size() < enough(d+1,vars.length)) {
			if (LoopBoundsAnalyzer.DEBUG)
				System.out.println("Not enough nodes!");
			return false;
		}
		
		//create HashMap which maps the unique values of the first variable
		//to separated lists of nodes
		HashMap<Integer, List<int[]>> unique = new HashMap<Integer, List<int[]>>();
		for (int i = 0; i < nodes.size(); i++) {
			int[] node = nodes.get(i);
			Integer firstVar = new Integer(node[0]);
			List<int[]> lst = unique.get(firstVar);
			if (lst == null) {
				lst = new ArrayList<int[]>();
				unique.put(firstVar, lst);
			}
			if (LoopBoundsAnalyzer.DEBUG)
				System.out.println("Node: " + Arrays.toString(node));
			lst.add(Arrays.copyOfRange(node, 1, node.length));
		}
		
		//Now check that, in our HashMap, we have lists of minimum sizes d+1, ..., 1
		boolean[] present = new boolean[triangleSize];
		Arrays.fill(present, false);
		for (List<int[]> hyperplane : unique.values()) {
			
			int size = Math.min(hyperplane.size(), triangleSize);
			
			/* while size is already present, decrease it, because hyperplane
			   can also serve lower size */
			while (size>0) {
				if (present[size-1] == false) {
					
					/* Check that hyperplane is in NCA for traingleSize=size */
					if (inNCA(size, hyperplane, Arrays.copyOfRange(vars, 1, vars.length))) {
						present[size-1] = true;
						break;
					}
				}
				size--;
			}
		}
	
		//If a hyperplane of some size is not present in NCA config, these nodes don't suffice
		for (boolean b : present) {
			if (b == false) {
				if (LoopBoundsAnalyzer.DEBUG)
					System.out.println("inNCA returning false!");
				return false;
			}
		}
		
		if (LoopBoundsAnalyzer.DEBUG)
			System.out.println("inNCA returning true!");
		return true;
	}
	
	/** In most cases, we find more test-nodes than necessary.
		The matrix to solve by PolynoomGenerator is then not a square.
		Therefore the solution/coefficients it finds are least squares,
		not round integers.

		To solve this and to increase efficiency, this function reduces the set
		of test nodes to the minimum.
		*/
	private void reduceToSquare() {
		nodes = reduceToSquareAux(d+1, nodes, argNames);
	}
	
	private List<int[]> reduceToSquareAux(int desiredSize, List<int[]> nodes, String[] vars) {
		
		//if there is only one variable/dimension left, then there are enough
		//nodes if there are as least as many nodes as the triangleSize
		if (vars.length == 1) {
			if (LoopBoundsAnalyzer.DEBUG)
				System.out.println("[reduceToSquareAux] Return minimal list: " + nodes.subList(0, desiredSize));
			return nodes.subList(0, desiredSize);
		}
		
		//create HashMap which maps the unique values of the first variable
		//to separated lists of nodes
		HashMap<Integer, List<int[]>> unique = new HashMap<Integer, List<int[]>>();
		for (int i = 0; i < nodes.size(); i++) {
			int[] node = nodes.get(i);
			Integer firstVar = new Integer(node[0]);
			List<int[]> lst = unique.get(firstVar);
			if (lst == null) {
				lst = new ArrayList<int[]>();
				unique.put(firstVar, lst);
			}
			if (LoopBoundsAnalyzer.DEBUG)
				System.out.println("[reduceToSquareAux] Node: " + Arrays.toString(node));
			lst.add(Arrays.copyOfRange(node, 1, node.length));
		}
		
		//Now decrease the nodes to the minimal NCA configuration
		boolean[] done = new boolean[desiredSize];
		Arrays.fill(done, false);
		List<int[]> newnodes = new ArrayList<int[]>();
		
		for (Integer key : unique.keySet()) {
			List<int[]> hyperplane = unique.get(key);
			
			int size = Math.min(hyperplane.size(), desiredSize);
			
			/* while size is already present, decrease it, because hyperplane
			   can also serve lower size */
			while (size>0) {
				if (done[size-1] == false) {
					
					/* Check that hyperplane is in NCA for traingleSize=size */
					if (inNCA(size, hyperplane, Arrays.copyOfRange(vars, 1, vars.length))) {
						List<int[]> newhyperplane = reduceToSquareAux(size, hyperplane, Arrays.copyOfRange(vars, 1, vars.length));
						
						//generate nodes from the first variables and all endings
						for (int[] nodeEnd : newhyperplane) {
							int[] node = new int[nodeEnd.length+1];
							node[0] = key.intValue();
							for (int i = 0; i < nodeEnd.length; i++) {
								node[i+1] = nodeEnd[i];
							}
							newnodes.add(node);
							if (LoopBoundsAnalyzer.DEBUG)
								System.out.println("[reduceToSquareAux] Added node to minimal list: " + Arrays.toString(node));
						}
						
						done[size-1] = true;
						break;
					}
				}
				size--;
			}
		}
		return newnodes;
	}
	
	/** Calculates the minimum number of nodes needed for interpolation,
	 * for a given degree 'deg' and number of variables / dimension 'dim'.
	 * 
	 * @param deg The expected degree
	 * @param dim The number of variables / dimension
	 * @return The minimum number of nodes needed for interpolation
	 */
	private int enough(int deg, int dim) {
		int result = 0;
		if (dim > 0) {
			for (int i = 1; i <= deg; i++) {
				result += enough(i, dim-1);
			}
		} else {
			result = 1;
		}
		return result;
	}
}
