package loopbounds.parsetree;

import loopbounds.Polynomial;
import loopbounds.PiecewisePolynomial;

import java.util.List;

import org.jmlspecs.openjml.JmlToken;
import org.jmlspecs.openjml.JmlTree;
import org.jmlspecs.openjml.JmlTree.JmlMethodClause;
import org.jmlspecs.openjml.JmlTree.JmlMethodDecl;
import org.jmlspecs.openjml.JmlTree.JmlSpecificationCase;
import org.jmlspecs.openjml.JmlTree.JmlStatementLoop;

import com.sun.tools.javac.tree.JCTree.JCExpression;
import com.sun.tools.javac.tree.JCTree.JCStatement;
import com.sun.tools.javac.util.Context;

/**
 * Class for adding a loop bound to the parse tree, as a JML annotation. Class
 * is intended to be used though one of the static 'annotate' methods. Annotates both
 * the loop and the method, in order to be verified by the KeY tool.
 * 
 * @author Rody Kersten
 * 
 */
public abstract class AnnotationGenerator {

	/**
	 * Annotates a method and the loop inside this method with the appropriate
	 * JML annotations to verify a loop bound of the given polynomial.
	 * 
	 * Our upper bound can be a polynomial over rationals. This can however not
	 * be expressed in JML in a straight-forward fashion. The solution is to
	 * write the polynomial p with rational coefficients as q/step. Then, q is a
	 * polynomial over integers and becomes our decreases clause. We then have
	 * to add additional program variables and an invariant in order to enable
	 * verification of the fact that step is correct. To this end, we add for
	 * each variable v, an auxiliary variable v_old, which stores the value of v
	 * before executing the loop body (the JML predicate \old() refers to the
	 * value of v before entering the method). For each iteration except the
	 * first, the value of v must then be equal to v_old+step.
	 * 
	 * Plus, the postcondition 'true'.
	 * 
	 * @param decreases
	 *            The decreases clause
	 * @param assignable
	 * 			  The assignable program variables
	 * @param meth
	 *            The method to annotate, containing only a single while loop
	 *            (so must be custom made)
	 * @param loop
	 *            The while loop inside 'meth'
	 * @param context
	 *            The context from the OpenJML API
	 */
	public static boolean annotate(JCExpression decreases, List<String> assignable,
			JmlMethodDecl meth,	JCStatement loop, Context context) {
		
		//by default, do not overwrite existing decreases clauses and write
		//annotations that KeY can verify
		return annotate(decreases, assignable, meth, loop, context, false, false);
	}
	
	/**
	 * Annotates a method and the loop inside this method with the appropriate
	 * JML annotations to verify a loop bound of the given polynomial.
	 * 
	 * Our upper bound can be a polynomial over rationals. This can however not
	 * be expressed in JML in a straight-forward fashion. The solution is to
	 * write the polynomial p with rational coefficients as q/step. Then, q is a
	 * polynomial over integers and becomes our decreases clause. We then have
	 * to add additional program variables and an invariant in order to enable
	 * verification of the fact that step is correct. To this end, we add for
	 * each variable v, an auxiliary variable v_old, which stores the value of v
	 * before executing the loop body (the JML predicate \old() refers to the
	 * value of v before entering the method). For each iteration except the
	 * first, the value of v must then be equal to v_old+step.
	 * 
	 * Plus, the postcondition 'true'.
	 * 
	 * @param decreases
	 *            The decreases clause
	 * @param assignable
	 * 			  The assignable program variables
	 * @param meth
	 *            The method to annotate, containing only a single while loop
	 *            (so must be custom made)
	 * @param loop
	 *            The while loop inside 'meth'
	 * @param context
	 * 			  The OpenJML context
	 * @param correct
	 * 			  If true, write JML annotations that are correct with respect to the
	 * 			  JML specification, rather than annotations that can be verified by KeY.
	 * @param overwriteDecreases
	 * 			  If true, existing decreases clauses are overwritten
	 * @param context
	 *            The context from the OpenJML API
	 */
	public static boolean annotate(JCExpression decreases, List<String> assignable, JmlMethodDecl meth,
			JCStatement loop, Context context, boolean correct, boolean overwriteDecreases) {
		
		if (annotateLoop(decreases, assignable, loop, context, correct, overwriteDecreases)) {
			//only annotate the method if the loop has also been annotated
			annotateMethod(meth, context, correct);
			return true;
		} 
		return false;
	}
	
	/**
	 * See first annotate method. Annotates a loop with a polynomial LBF.
	 * 
	 * @param p
	 *            The polynomial LBF
	 * @param meth
	 *            The method to annotate, containing only a single while loop
	 *            (so must be custom made)
	 * @param loop
	 *            The while loop inside 'meth'
	 * @param context
	 *            The context from the OpenJML API
	 * @param correct
	 * 			  If true, write JML annotations that are correct with respect to the
	 * 			  JML specification, rather than annotations that can be verified by KeY.
	 * @param overwriteDecreases
	 * 			  If true, existing decreases clauses are overwritten
	 */
	public static boolean annotate(Polynomial p, JmlMethodDecl meth,
			JCStatement loop, Context context, boolean correct, boolean overwriteDecreases) {
		return annotate(
				p.toJCExpression(context),
				p.getVariables().getFullNames(),
				meth,
				loop,
				context,
				correct,
				overwriteDecreases
		);
	}
	
	/**
	 * See first annotate method. Annotates a loop with a polynomial LBF.
	 * 
	 * @param p
	 *            The polynomial LBF
	 * @param meth
	 *            The method to annotate, containing only a single while loop
	 *            (so must be custom made)
	 * @param loop
	 *            The while loop inside 'meth'
	 * @param context
	 *            The context from the OpenJML API
	 */
	public static boolean annotate(Polynomial p, JmlMethodDecl meth,
			JCStatement loop, Context context) {
		return annotate(
				p.toJCExpression(context),
				p.getVariables().getFullNames(),
				meth,
				loop,
				context
		);
	}
	
	/**
	 * See first annotate method. Annotates a loop with a piecewise polynomial LBF.
	 * 
	 * @param lbf
	 *            The piecewise polynomial LBF
	 * @param meth
	 *            The method to annotate, containing only a single while loop
	 *            (so must be custom made)
	 * @param loop
	 *            The while loop inside 'meth'
	 * @param context
	 *            The context from the OpenJML API
	 * @param correct
	 * 			  If true, write JML annotations that are correct with respect to the
	 * 			  JML specification, rather than annotations that can be verified by KeY.
	 * @param overwriteDecreases
	 * 			  If true, existing decreases clauses are overwritten
	 */
	public static boolean annotate(PiecewisePolynomial lbf, JmlMethodDecl meth,
			JCStatement loop, Context context, boolean correct, boolean overwriteDecreases) {
		return annotate(
				lbf.asDecreasesClause(context),
				lbf.getUniqueVariables(),
				meth,
				loop,
				context,
				correct,
				overwriteDecreases
		);
	}
	
	/**
	 * See first annotate method. Annotates a loop with a piecewise polynomial LBF.
	 * 
	 * @param lbf
	 *            The piecewise polynomial LBF
	 * @param meth
	 *            The method to annotate, containing only a single while loop
	 *            (so must be custom made)
	 * @param loop
	 *            The while loop inside 'meth'
	 * @param context
	 *            The context from the OpenJML API
	 */
	public static boolean annotate(PiecewisePolynomial lbf, JmlMethodDecl meth,
			JCStatement loop, Context context) {
		return annotate(
				lbf.asDecreasesClause(context),
				lbf.getUniqueVariables(),
				meth,
				loop,
				context
		);
	}
	
	/** Adds a postcondition of 'true', if no postcondition exists.
	 * And, if correct annotations are used, adds 'assignable \everything'.
	 **/
	private static void annotateMethod(JmlMethodDecl meth, Context context, boolean correct) {

		JmlTree.Maker M = JmlTree.Maker.instance(context);
		
		// Method postcondition = true or existing postcondition
		JCExpression post = M.Ident("true");

		// Construct method annotation
		JmlMethodClause mcpost = M.JmlMethodClauseExpr(JmlToken.ENSURES, post);
		JmlMethodClause mcassign = M.JmlMethodClauseExpr(JmlToken.ASSIGNABLE, M.JmlSingleton(JmlToken.BSEVERYTHING));
		com.sun.tools.javac.util.List<JmlMethodClause> clausesPost = com.sun.tools.javac.util.List
				.<JmlMethodClause> of(mcpost);
		com.sun.tools.javac.util.List<JmlMethodClause> clausesAssign = com.sun.tools.javac.util.List
		.<JmlMethodClause> of(mcassign);
		com.sun.tools.javac.util.List<JmlMethodClause> clausesPostAndAssign = com.sun.tools.javac.util.List
		.<JmlMethodClause> of(mcpost, mcassign);
		JmlSpecificationCase jmlc;
		if (correct)
			jmlc = M.JmlSpecificationCase(null, false, null, null, clausesPostAndAssign);
		else
			jmlc = M.JmlSpecificationCase(null, false, null, null, clausesPost);
		com.sun.tools.javac.util.List<JmlSpecificationCase> cases = com.sun.tools.javac.util.List
				.<JmlSpecificationCase> of(jmlc);
		
		if (meth.cases == null) {
			meth.cases = M.JmlMethodSpecs(cases);
		} else if (meth.cases.cases == null) {
			meth.cases.cases = cases;
		} else {
			
			//For each specification case, add 'ensures true' if no postcondition exists
			com.sun.tools.javac.util.List<JmlSpecificationCase> lst = meth.cases.cases;
			if (lst.head != null && lst.head.clauses != null) {
				com.sun.tools.javac.util.List<JmlMethodClause> lst2 = lst.head.clauses;
				do {
					if (lst2.head != null && lst2.head.token != null) {
						
						//find postcondition
						boolean foundPost = false, foundAssign = false;
						do {
							if (lst2.head.token == JmlToken.ENSURES)
								foundPost = true;
							else if (lst2.head.token == JmlToken.ASSIGNABLE)
								foundAssign = true;
							lst2 = lst2.tail;
						} while (lst2.tail != null);
						
						//if none exists, add postcondition
						if (!foundPost)
							lst.head.clauses = lst.head.clauses.appendList(clausesPost);
						
						//if this is a lightweight specification and no assignables clause
						//exists yet, add 'assignable \everything'
						boolean lightWeightSpec = lst.head.token == null;
						if (lightWeightSpec && !foundAssign && correct)
							lst.head.clauses = lst.head.clauses.appendList(clausesAssign);
					}
				lst = lst.tail;
				} while (lst != null && lst.head != null && lst.head.clauses != null);
			}
		}
	}
	
	/** Write a loop annotation.
	 * 
	 * First checks if decreases, assignable and/or invariant annotations exist;
	 * 
	 * If a decreases clause already exists, either overwrites it or returns false, depending
	 * on the 'overwriteDecreases' parameter.
	 * 
	 * If assignable clauses already exist, keep them, don't add anything (i.e. assume that
	 * they are correct).
	 * 
	 * If a loop invariant exists, keep it. Else add the loop invariant 'true' (KeY needs this).
	 * 
	 * @param decreases The decreases clause to annotate the loop with
	 * @param assignable The assignable variables, in case assignable annotations do
	 * not already exist
	 * @param loop The loop to annotate
	 * @param context The OpenJML API context
	 * @param correct
	 * 			  If true, write JML annotations that are correct with respect to the
	 * 			  JML specification, rather than annotations that can be verified by KeY.
	 * @param overwriteDecreases Whether or not to overwrite an existing decreases clause
	 * @return True, if an annotation has been added. The single case where this does not happen
	 * is when a decreases clause already exists and 'overwriteDecreases' = false.
	 */
	private static boolean annotateLoop(JCExpression decreases, List<String> assignable,
			JCStatement loop, Context context, boolean correct, boolean overwriteDecreases) {
		
		JmlTree.Maker M = JmlTree.Maker.instance(context);
		
		com.sun.tools.javac.util.List<JmlTree.JmlStatementLoop> specs =
			ParseTreeWrapper.getLoopSpecs(loop);
		
		//Check if loop contains invariant, decreases and/or assignable clauses
		boolean containsInvariant = false;
		boolean containsAssignables = false;
		boolean containsDecreases = false;
		for (int i = 0; i < specs.size(); i++) {
			if (specs.get(i).token == JmlToken.LOOP_INVARIANT)
				containsInvariant = true;
			else if (specs.get(i).token == JmlToken.ASSIGNABLE)
				containsAssignables = true;
			else if (specs.get(i).token == JmlToken.DECREASES)
				containsDecreases = true;
		}

		if (!containsAssignables && !correct) {
			// Add assignable clauses for all variables
			for (String var : assignable) {
				JmlStatementLoop assorig = M.JmlStatementLoop(JmlToken.ASSIGNABLE, M.Ident(var.replaceAll("\\[.*\\]","[*]")));
				specs = specs.append(assorig);
			}
		}
		
		// Add invariant 'true' if none exists
		if (!containsInvariant)
			specs = specs.append(M.JmlStatementLoop(JmlToken.LOOP_INVARIANT, M.Literal(true)));


		// Decreases
		JmlStatementLoop decr = M.JmlStatementLoop(JmlToken.DECREASES, decreases);
		if (!containsDecreases) {
			specs = specs.append(decr);
		} else if (overwriteDecreases) {
			for (int i = 0; i < specs.size(); i++) {
				if (specs.get(i).token == JmlToken.DECREASES) {
					//com.sun.tools.javac.util.List does not provide a set() method
					JmlStatementLoop specArray[] =
						specs.toArray(new JmlStatementLoop[specs.size()]);
					specArray[i] = decr;
					specs = com.sun.tools.javac.util.List.<JmlTree.JmlStatementLoop>from(specArray);
				}
			}
		} else {
			//Decreases clause exists that may not be overwritten
			return false;
		}
		
		ParseTreeWrapper.setLoopSpecs(loop, specs);
		return true;
	}
}
