package loopbounds.parsetree;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTree;
import org.jmlspecs.openjml.JmlTreeScanner;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCBlock;
import com.sun.tools.javac.tree.JCTree.JCStatement;
import com.sun.tools.javac.tree.JCTree.JCWhileLoop;
import com.sun.tools.javac.tree.JCTree.*;
import com.sun.tools.javac.util.Context;


/**
 * The main class for the branch splitting.
 * 
 * @author Robin Oostrum
 * 
 */
public class IfSplitter extends JmlTreeScanner implements IJmlVisitor {
	
	/**
	 * Recursively splits all if- and switch-statements within a while-, for-, or do-while-loop, and returns
	 * multiple loops with the split branches, or just the original loop if no splitting is possible.
	 * @param loop the loop that has to be split
	 * @param context the compiler used within the Loopbounds environment
	 * @return a list of loops, in which all branches have been split.
	 */
	public static List<JCStatement> split(JCStatement loop, Context context) {
		
		JmlTree.Maker make = JmlTree.Maker.instance(context);
		List<JCStatement> ret = new ArrayList<JCStatement>();
		
		JCStatement LoopBody = null;
		JCExpression condition = null;
		List<JCStatement> init = null;
		List<JCExpressionStatement> step = null;
		
		if (!loop.toString().contains("if") && !loop.toString().contains("switch")) {
			ret.add(loop);
			return ret;
		}
		
		if (loop instanceof JCWhileLoop) {
			LoopBody = ((JCWhileLoop) loop).getStatement();
			condition = ((JCWhileLoop) loop).getCondition();
		} else if (loop instanceof JCForLoop) {
			LoopBody = ((JCForLoop) loop).getStatement();
			condition = ((JCForLoop) loop).getCondition();
			init = ((JCForLoop) loop).getInitializer();
			step = ((JCForLoop) loop).getUpdate();
		} else if (loop instanceof JCDoWhileLoop) {
			LoopBody = ((JCDoWhileLoop) loop).getStatement();
			condition = ((JCDoWhileLoop) loop).getCondition();
		} else {
			ret.add(loop);
			return ret;
		}
		
		if (!(LoopBody instanceof JCBlock)) {
			ret.add(loop);
			return ret;
		}
		
		// create different loop bodies for each branch splitting
		@SuppressWarnings("unchecked")
		ArrayList<JCStatement>[] bodies = (ArrayList<JCStatement>[])new ArrayList[5]; 
		bodies[0] = new ArrayList<JCStatement>();
		List<JCStatement> allstats = ((JCBlock) LoopBody).getStatements();
		long flags = ((JCBlock) LoopBody).flags;
		boolean splitFlag = false;	// flag raised when a split occurs
		for (JCStatement check : allstats) { 
			if (!splittable(check) || splitFlag) {
				for (int i = 1; i < bodies.length; i++) {
					if (bodies[i] != null)
						bodies[i].add(check);
				}
				bodies[0].add(check);
			}
			else if (check.toString().startsWith("{") && check.toString().endsWith("}") &&
					check.toString().contains("if")) {
				for (int i = 1; i < bodies.length; i++) {
					if (bodies[i] != null)
						bodies[i].addAll(((JCBlock) check).getStatements());
				}
				bodies[0].addAll(((JCBlock) check).getStatements());
			}
			else {
				splitFlag = true;
				List<JCStatement> splitted = splitStatement(check);
				for (int i = 0; i < splitted.size(); i++) {
					if (bodies[i] == null && i>0 && bodies[i-1] != null) {
						bodies[i] = new ArrayList<JCStatement>();
						for (int j=0; j<bodies[i-1].size()-1; j++) {
							bodies[i].add(bodies[i-1].get(j));
						}
					}
					if (splitted.get(i) != null)
						bodies[i].add(splitted.get(i));
				}
			}

	    }
		
		// recreate the loops with the new loop bodies
		for (int i=0; i<bodies.length; i++) {
			if (bodies[i] != null) {
				JCStatement []body_i = new JCStatement[bodies[i].size()];
				for (int j=0; j<bodies[i].size(); j++) {			
					body_i[j] = bodies[i].get(j);
					
					// There may, after branch splitting, be a return statement
					// in the middle of the loop. Stop there. Effectively, this removes
					// all (unreachable) code after the return statement.
					if (body_i[j] instanceof JCReturn) {
						body_i = Arrays.copyOf(body_i, j+1);
						break;
					}
				}
				com.sun.tools.javac.util.List<JCStatement> stats = com.sun.tools.javac.util.List.from(body_i);
				JCBlock body = make.Block(flags, stats);
				if (loop instanceof JCWhileLoop) {
		        	JCWhileLoop jcwl = make.WhileLoop(condition,body);
		        	JmlTree.JmlWhileLoop WL = make.JmlWhileLoop(jcwl,null);
					ret.addAll(split(WL, context));
				}
				else if (loop instanceof JCForLoop) {
					JCForLoop jcfl = make.ForLoop((com.sun.tools.javac.util.List<JCStatement>) init,
							condition, (com.sun.tools.javac.util.List<JCExpressionStatement>) step,body);
					JmlTree.JmlForLoop FL = make.JmlForLoop(jcfl, null);
					ret.addAll(split(FL, context));
					}
				else if (loop instanceof JCDoWhileLoop) {
					JCDoWhileLoop jcdwl = make.DoLoop(body, condition);
					JmlTree.JmlDoWhileLoop DWL = make.JmlDoWhileLoop(jcdwl, null);
					ret.addAll(split(DWL, context));
				}
			}
		}

		return ret;
	}
	
	/**
	 * Splits a given statement of type {@link JCStatement} into branches, until each branch cannot be
	 * split any further.
	 * @param s the statement that has to be split
	 * @return a list of all branches after splitting
	 */
	private static List<JCStatement> splitStatement(JCStatement s) {
		List<JCStatement> splitted = new ArrayList<JCStatement>();
		//System.out.println("Statement to split: " + s.toString());
		// split if's
		if (s instanceof JCIf) {      		
			JCStatement thenpart = ((JCIf) s).getThenStatement();
			JCStatement	elsepart = ((JCIf) s).getElseStatement();

			splitted.add(thenpart);
			splitted.add(elsepart);
    	}
		
		// split switches
    	else if (s instanceof JCSwitch) {
    		List<JCTree.JCCase> s_cases = ((JCSwitch) s).getCases();
    		
    		for (JCTree.JCCase c : s_cases) {
    			List<JCStatement> stats = c.getStatements();
    			if (stats != null)
    				splitted.addAll(stats);
    		}
    	}
		
		return splitted;

	}

	/**
	 * Tests whether a given statement of type {@link JCStatement} is splittable into branches.
	 * @param s the statement that has to be checked for splittability
	 * @return true iff s is splittable
	 */
	private static boolean splittable(JCStatement s) {
		if (s instanceof JCIf)
			return true;
		if (s instanceof JCSwitch)
			return true;
		if (s != null && s.toString().startsWith("{") && s.toString().endsWith("}") &&
				s.toString().contains("if"))
			return true;
		
		return false;
	}

	@Override
	public void visitIf (JCIf that) {
		super.visitIf(that);
	}
}
