package loopbounds.parsetree;

import java.util.List;

import loopbounds.VarInfo;

import org.jmlspecs.openjml.JmlTreeScanner;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCAssign;
import com.sun.tools.javac.tree.JCTree.JCAssignOp;
import com.sun.tools.javac.tree.JCTree.JCBreak;
import com.sun.tools.javac.tree.JCTree.JCContinue;
import com.sun.tools.javac.tree.JCTree.JCExpression;
import com.sun.tools.javac.tree.JCTree.JCIdent;
import com.sun.tools.javac.tree.JCTree.JCReturn;
import com.sun.tools.javac.tree.JCTree.JCSkip;
import com.sun.tools.javac.tree.JCTree.JCStatement;
import com.sun.tools.javac.tree.JCTree.JCUnary;
import com.sun.tools.javac.tree.TreeMaker;

/**
 * Very simple program slicer. Slices the whole loop body if the variables in the condition are not assigned.
 * 
 * @author Wouter Smeenk
 * 
 */
public class Slicer extends JmlTreeScanner {
    private final List<String> vars;
    private boolean assigned;

    private Slicer(List<String> vars) {
        this.vars = vars;
        this.assigned = false;
    }

    /**
     * Slices the whole loop body if the variables in the condition are not assigned.
     * 
     * @param loop The loop to slice
     * @param code The code
     * @param classIndex The class index of the method the loop is in.
     * @param methIndex The method index where the loop is in.
     * @return Whether the loop was sliced or not.
     */
    public static boolean slice(JCStatement loop, ParseTreeWrapper code,
            int classIndex, int methIndex) throws UnsupportedTypeException,
            UndeclaredVariableException {
        JCExpression cond = ParseTreeWrapper.getLoopCondition(loop);
        VarInfo varInfo = VariableScanner.getUndeclaredVars(cond);
        varInfo.loadTypes(code, classIndex, methIndex);
        for (Class<?> type : varInfo.getTypes()) {
            // If we have a non primitive or array dont slice (is array a
            // primitive?)
            if (!type.isPrimitive() | type.isArray()) {
                return false;
            }
        }
        Slicer slicer = new Slicer(varInfo.getFullNames());
        // If the variables in the condition are not assigned we slice the body away.
        if (!slicer.isAssigned(ParseTreeWrapper.getLoopBody(loop))) {
            JCSkip skip = TreeMaker.instance(code.context()).Skip();
            ParseTreeWrapper.setLoopBody(loop, skip);
            return true;
        } else {
            return false;
        }
    }

    private boolean isAssigned(JCStatement statement) {
        assigned = false;
        scan(statement);
        return assigned;
    }

    private void checkVars(JCExpression lhs) {
        if (lhs instanceof JCTree.JCIdent) {
            JCIdent ident = (JCIdent) lhs;
            if (vars.contains(ident.name.toString())) {
                assigned = true;
            }
        }
    }

    @Override
    public void visitAssign(JCAssign tree) {
        checkVars(tree.lhs);
        super.visitAssign(tree);
    }

    @Override
    public void visitAssignop(JCAssignOp tree) {
        checkVars(tree.lhs);
        super.visitAssignop(tree);
    }
    
    @Override
    public void visitBreak(JCBreak tree) {
        assigned = true;
    }
    
    @Override
    public void visitReturn(JCReturn tree) {
        assigned = true;
    }
    
    @Override
    public void visitContinue(JCContinue tree) {
        assigned = true;
    }

    @Override
    public void visitUnary(JCUnary tree) {
        // Only pre- and postfix increment and decrement assign the variable
        switch (tree.getKind()) {
        case PREFIX_DECREMENT:
            checkVars(tree.arg);
            break;
        case PREFIX_INCREMENT:
            checkVars(tree.arg);
            break;
        case POSTFIX_DECREMENT:
            checkVars(tree.arg);
            break;
        case POSTFIX_INCREMENT:
            checkVars(tree.arg);
            break;
        default:
            break;
        }
        super.visitUnary(tree);
    }

}
