package loopbounds.parsetree;

import loopbounds.LoopBoundsAnalyzer;

import java.util.ArrayList;
import java.util.List;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;
import org.jmlspecs.openjml.JmlTree.JmlDoWhileLoop;
import org.jmlspecs.openjml.JmlTree.JmlForLoop;
import org.jmlspecs.openjml.JmlTree.JmlWhileLoop;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCAssign;
import com.sun.tools.javac.tree.JCTree.JCAssignOp;
import com.sun.tools.javac.tree.JCTree.JCBinary;
import com.sun.tools.javac.tree.JCTree.JCIdent;
import com.sun.tools.javac.tree.JCTree.JCLiteral;
import com.sun.tools.javac.tree.JCTree.JCStatement;
import com.sun.tools.javac.tree.JCTree.JCUnary;

/** Class for finding the step, based on syntax.
 * 
 * Looks for statements like i++ or i -= 5, to determine a hypothetical step (in these
 * cases, 1 and 5, respectively).
 * 
 * @author Rody Kersten
 *
 */
public class StepFinder extends JmlTreeScanner implements IJmlVisitor {

	private int step;
	private double trust;
	private List<String> varsInCondition;
	private boolean inCondition, inForUpdate;

	/** Constructor */
	private StepFinder() {
		super();
		this.step = 0;
		this.trust = 0.0;
		this.varsInCondition = new ArrayList<String>();
		this.inCondition = false;
		this.inForUpdate = false;
	}

	/**
	 * Try to find the step/increment in a loop.
	 * 
	 * @param loop
	 *            The OpenJml parse tree of the loop.
	 * @return The found step, or 0 if the step cannot be found.
	 */
	public static int getStep(JCStatement loop) {
		if (LoopBoundsAnalyzer.DEBUG)
			System.out.println("Trying to find step!");
		StepFinder sf = new StepFinder();
		sf.scan(loop);
		if (LoopBoundsAnalyzer.DEBUG) {
			if (sf.step > 0) {
				System.out.println("Found step: " + sf.step);
			} else {
				System.out.println("Unable to find step, set to 1");
			}
		}
		return sf.step > 0 ? sf.step : 1;
	}

	@Override
	public void visitJmlWhileLoop(JmlWhileLoop tree) {
		inCondition = true;
		scan(tree.cond);
		inCondition = false;

		scan(tree.body);
	}

	@Override
	public void visitJmlForLoop(JmlForLoop tree) {
		inCondition = true;
		scan(tree.cond);
		inCondition = false;

		inForUpdate = true;
		scan(tree.step);
		inForUpdate = false;

		scan(tree.body);
	}

	@Override
	public void visitJmlDoWhileLoop(JmlDoWhileLoop tree) {
		inCondition = true;
		scan(tree.cond);
		inCondition = false;

		scan(tree.body);
	}

	@Override
	public void visitAssign(JCAssign tree) {
		try {
			JCIdent var = (JCIdent) tree.lhs;
			if (varsInCondition.contains(var.name.toString())) {
				// see if rhs is this var +/- a constant
				JCBinary bin = (JCBinary) tree.rhs;
				if (!(bin.getTag() == JCTree.PLUS || bin.getTag() == JCTree.MINUS)) {
					return;
				}
				if ((bin.lhs instanceof JCIdent)
						&& (((JCIdent) bin.lhs).name.equals(var.name))) {
					JCLiteral lit = (JCLiteral) bin.rhs;
					Integer stepInteger = (Integer) lit.value;
					if (inForUpdate) {
						this.step = stepInteger.intValue();
						this.trust = 0.9;
					} else if (this.trust <= 0.5) {
						this.step = stepInteger.intValue();
						this.trust = 0.5;
					}
				} else if ((bin.rhs instanceof JCIdent)
						&& (((JCIdent) bin.rhs).name.equals(var.name))) {
					JCLiteral lit = (JCLiteral) bin.lhs;
					Integer stepInteger = (Integer) lit.value;
					if (inForUpdate) {
						this.step = stepInteger.intValue();
						this.trust = 0.9;
					} else if (this.trust <= 0.5) {
						this.step = stepInteger.intValue();
						this.trust = 0.5;
					}
				}
			}
		} catch (ClassCastException e) {
			return;
		}
	}

	@Override
	public void visitIdent(JCIdent tree) {
		if (inCondition && !varsInCondition.contains(tree.name.toString()))
			varsInCondition.add(tree.name.toString());
	}

	@Override
	public void visitAssignop(JCAssignOp tree) {
		if (tree.getTag() != JCTree.PLUS_ASG
				&& tree.getTag() != JCTree.MINUS_ASG)
			return;

		try {
			JCIdent var = (JCIdent) tree.lhs;
			if (varsInCondition.contains(var.name.toString())) {
				// see if rhs is a constant
				JCLiteral lit = (JCLiteral) tree.rhs;
				Integer stepInteger = (Integer) lit.value;

				if (inForUpdate) {
					this.step = stepInteger.intValue();
					this.trust = 0.9;
				} else if (this.trust <= 0.5) {
					this.step = stepInteger.intValue();
					this.trust = 0.5;
				}
			}
		} catch (ClassCastException e) {
			return;
		}
	}

	@Override
	public void visitUnary(JCUnary tree) {
		if (tree.getTag() == JCTree.POSTINC || tree.getTag() != JCTree.PREINC
				|| tree.getTag() == JCTree.POSTDEC
				|| tree.getTag() != JCTree.PREDEC) {
			try {
				JCIdent var = (JCIdent) tree.arg;
				if (varsInCondition.contains(var.name.toString())) {
					if (inForUpdate) {
						this.step = 1;
						this.trust = 0.9;
					} else if (this.trust <= 0.5) {
						this.step = 1;
						this.trust = 0.5;
					}
				}
			} catch (ClassCastException e) {
				return;
			}
		}
	}
}