package loopbounds.parsetree;

import loopbounds.LoopBoundsAnalyzer;
import java.lang.reflect.Array;
import java.util.ArrayList;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTree.JmlMethodDecl;
import org.jmlspecs.openjml.JmlTree.JmlVariableDecl;
import org.jmlspecs.openjml.JmlTreeScanner;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.JCTree.JCArrayTypeTree;
import com.sun.tools.javac.tree.JCTree.JCExpression;

/**
 * Class for finding the type of a variable in a parse tree.
 * <P>
 * The class extends JmlTreeScanner and uses the visitor pattern.
 * 
 * @author Rody Kersten
 *
 */
public class TypeScanner extends JmlTreeScanner implements IJmlVisitor {
	
	private String var;
	private ArrayList<String> varSplit;
	private JCExpression type;
	private JCTree meth;
	
	private TypeScanner(String var) {
		this.var = var;
		splitVar();
	}

	private TypeScanner(String var, JCTree meth) {
		this.var = var;
		splitVar();
		this.meth = meth;
	}
	
	/**
	 * Split the variable into appropriate class parts
	 */
	private void splitVar(){
		this.varSplit = new ArrayList<String>();
		if (var.contains(".")) {
			String[] splitVar = var.split("\\."); // Note: . matches any character, \\. escapes .
			for(int i = 0; i < splitVar.length; i++){
				varSplit.add(splitVar[i]);
			}
		}
	}
	
	/** Scan the JCTree 'tree' for the type of variable 'var'.
	 * Walks through the parse tree, looking for a declaration of variable 'var'.
	 * 
	 * @param var The variable we are looking for
	 * @param tree The parse tree in which we are looking
	 * @return The type of the first encountered declaration of variable 'var'
	 */
	public static Class<?> getType(String var, JCTree tree) throws UnsupportedTypeException {
		TypeScanner ts = new TypeScanner(var);
		ts.scan(tree);
		return toClass(ts.type);
	}
	
	/** Scan the JCTree 'tree' for the type of variable 'var'.
	 * Looks first in the method declaration 'meth'. Then, 
	 * if the variable isn't declared in the local/inner scope (which can
	 * override the outer scope), searches in the outer scope (fields
	 * of the class).
	 * 
	 * @param var The variable we are looking for
	 * @param meth The method in which we're looking (inner scope)
	 * @param cls The class in which we're looking (outer scope)
	 * @return The type of the first encountered declaration of variable 'var'
	 * @throws UndeclaredVariableException 
	 */
	public static Class<?> getType(String var, JCTree meth, JCTree cls, ArrayList<CodeClass> classes) throws UnsupportedTypeException, UndeclaredVariableException {	
		return toClass(getExpression(var, meth, cls, classes));
	}
	
	
	/** Similar to the getType method, but look recursively and determine the actual type for arrays and fields
	 * @param var Variable name to look for
	 * @param meth Code tree of method in with variable is found
	 * @param cls Code tree of class in with variable is found
	 * @param classes Variable definitions found in code
	 * @param intType
	 * @return JCExpression Of the variable type
	 * @throws UnsupportedTypeException
	 * @throws UndeclaredVariableException
	 */
	public static JCExpression getExpression(String var, JCTree meth, JCTree cls, ArrayList<CodeClass> classes) throws UnsupportedTypeException, UndeclaredVariableException {
		TypeScanner ts = new TypeScanner(var, meth);	
		// If we've got a field access process all the parts
		if (ts.varSplit.size() > 1) {
			JCExpression tp = null; 
			int i = 0;
			if (ts.varSplit.get(0).contentEquals("this")) {
				i++; // Forward past the "this" part if we have it.
			}
			JCExpression parent = ts.getVarType(ts.varSplit.get(i), meth, cls);
			for(i++; i < ts.varSplit.size(); i++) {
				if (parent instanceof JCArrayTypeTree && ts.varSplit.get(i).contentEquals("length")) {
					return ts.getVarType(var, meth, cls);
				}
				tp = ts.getChildType(parent, ts.varSplit.get(i), classes);
				parent = tp;
			}
			if (parent == null)
				throw new UndeclaredVariableException("Variable not found in class definitions");
			return parent;
		}
		if (var.contains("[") && var.contains("]")) {
		    // Remove the whole array index part
			var = var.substring(0,var.indexOf('['));
		}
		return ts.getVarType(var, meth, cls);
	}
	
	/**
	 * Returns the JCExpression defining the specified variable
	 * @param var The variable to find
	 * @param meth The method to looking
	 * @param cls The class to fallback to
	 * @return The expression containing the requested variable
	 */
	private JCExpression getVarType(String var, JCTree meth, JCTree cls) {
		this.var = var;
		scan(meth);
		if (type == null) {
			scan(cls);
		}
		return type;
	}
	
	/**
	 * Look through the precompiled list of classes and the variables and their types 
	 * to find the type of the variable
	 * @param parent The expression for the parent type
	 * @param child The name of the child to look for (the field)
	 * @param classes The list of classes
	 * @return The expression for the field
	 */
	private JCExpression getChildType(JCExpression parent, String child, ArrayList<CodeClass> classes){
		// Look through all the classes we know about and find our class (parent)
		for (int i = 0; i < classes.size(); i++) {
			if (classes.get(i).getName().equals(parent.toString())) { 
				// We found the right class, now look for the correct variable and return its type
				CodeClass current = classes.get(i);
				for (int j = 0; j < current.getFields().length; j++) {
					if (current.getFields()[j].equals(child)){ // We've found the type, return it.
						return current.getTypes()[j];
					}
				}
			}
		}
		return null;
	}
	
	/**
	 * Return the primitive type of a JCExpression
	 * @param type
	 * @return Class
	 * @throws UnsupportedTypeException
	 */
	public static Class<?> toClass(JCExpression type) throws UnsupportedTypeException {
		if (LoopBoundsAnalyzer.DEBUG)
			System.out.println("Looking up type: " + type);
		
		if (type.toString().equals("int")) {
			return int.class;
		}
		if (type.toString().equals("short")) {
			return short.class;
		}
		if (type.toString().equals("long")) {
			return long.class;
		}
		if (type.toString().equals("byte")) {
			return byte.class;
		}
		if (type.toString().equals("char")) {
			return char.class;
		}
		if (type.toString().equals("float")) {
			return float.class;
		}
		if (type.toString().equals("double")) {
			return double.class;
		} 
		if (type instanceof JCArrayTypeTree) {
			return Array.newInstance(toClass((JCExpression)((JCArrayTypeTree)type).elemtype), 1).getClass();
		} 
		try {
			return Class.forName(type.toString());
		} catch (ClassNotFoundException e) { } // Fall through
		
		throw new UnsupportedTypeException(type.toString());
	}
	
	@Override
	public void visitJmlVariableDecl (JmlVariableDecl tree) {
		String name = tree.name.toString();
		if (varSplit.contains(name)){
			type = tree.vartype;
		}
		if (name.equals(var)) {
			type = tree.vartype;
			if(LoopBoundsAnalyzer.DEBUG)
				System.out.println("Found VarDecl: " + var + ", type: " + tree.vartype);
		}
	}
	
	@Override
	public void visitJmlMethodDecl (JmlMethodDecl tree) {
		if (meth == null || meth.equals(tree)) {
			super.visitJmlMethodDecl(tree);
		}
	}
}