package loopbounds.parsetree;

import loopbounds.VarInfo;

import org.jmlspecs.openjml.IJmlVisitor;
import org.jmlspecs.openjml.JmlTreeScanner;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.TreeMaker;
import com.sun.tools.javac.tree.JCTree.*;
import com.sun.tools.javac.util.Names;
import com.sun.tools.javac.util.Context;

/** Class to replace all variables by their respective placeholders.
 * 
 *  For instance replaces an occurrence of this.a by this_a (if 'this_a' is the
 *  placeholder name determined by VarInfo's rename method).
 * 
 * @author Rody Kersten
 *
 */
public class VariableReplacer extends JmlTreeScanner implements IJmlVisitor {

	private VarInfo variables;
	private Context context;
	
	/** Replace all variables by their placeholders.
	 * For instance replaces an occurrence of this.a by this_a (if 'this_a' is the
	 * placeholder name determined by VarInfo's rename method).
	 * 
	 * @param variables VarInfo that contains the variables and their placeholders
	 * @param context The OpenJML context
	 */
	private VariableReplacer(VarInfo variables, Context context) {
		this.variables = variables;
		this.context = context;
	}
	
	/** Replace all variables by their placeholders.
	 * For instance replaces an occurrence of this.a by this_a (if 'this_a' is the
	 * placeholder name determined by VarInfo's rename method).
	 * 
	 * @param tree The JCTree to replace variables in
	 * @param variables VarInfo that contains the variables and their placeholders
	 * @param context The OpenJML context
	 */
	public static void replaceVariables(JCTree tree, VarInfo variables, Context context) {
		VariableReplacer far = new VariableReplacer(variables,context);
		far.scan(tree);
	}

	@Override
	public void visitVarDef(JCVariableDecl tree) {
		tree.init = (JCExpression) replace(tree.init);
		super.visitVarDef(tree);
	}
	
	@Override
    public void visitAssign(JCAssign tree) {
        tree.lhs = (JCExpression) replace(tree.lhs);
        tree.rhs = (JCExpression) replace(tree.rhs);
        super.visitAssign(tree);
    }
	
	@Override
    public void visitAssignop(JCAssignOp tree) {
        tree.lhs = (JCExpression) replace(tree.lhs);
        tree.rhs = (JCExpression) replace(tree.rhs);
        super.visitAssignop(tree);
    }
	
	@Override
    public void visitUnary(JCUnary tree) {
		tree.arg = (JCExpression) replace(tree.arg);
		super.visitUnary(tree);
    }
		
	@Override
	public void visitBinary(JCBinary tree) {
		tree.lhs = (JCExpression) replace(tree.lhs);
		tree.rhs = (JCExpression) replace(tree.rhs);
		super.visitBinary(tree);
	}
	
	@Override
    public void visitLetExpr(LetExpr tree) {
        scan(tree.defs);
        scan(tree.expr);
		tree.expr = (JCExpression) replace(tree.expr);
		super.visitLetExpr(tree);
    }
	
	/** Check if one of the replacement rules applies, then replace */
	private JCTree replace(JCTree tree) {
		if (tree instanceof JCFieldAccess)
			return replaceFieldAccess((JCFieldAccess) tree);
		else if (tree instanceof JCArrayAccess)
			return replaceArrayAccess((JCArrayAccess) tree);
		return tree;
	}
	
	/** Replaces a field access */
	private JCExpression replaceFieldAccess (JCFieldAccess field) {
		int index = variables.findFullName(field.toString());
		if (index >= 0) {
			TreeMaker make = TreeMaker.instance(context);
			Names makeName = Names.instance(context);
			return make.Ident(makeName.fromString(variables.getPlaceHolder(index)));
		}
		return field;
	}
	
	/** Replaces an array access */
	private JCExpression replaceArrayAccess (JCArrayAccess aa) {
		int index = variables.findFullName(aa.toString());
		if (index >= 0) {
			TreeMaker make = TreeMaker.instance(context);
			Names makeName = Names.instance(context);
			return make.Ident(makeName.fromString(variables.getPlaceHolder(index)));
		}
		return aa;
	}
}
